/*
**    TP 5ETI CPE Lyon
**    Copyright (C) 2011 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/



#include <GL/glut.h>


#include <opengl_drawer.hpp>
#include <texture.hpp>

namespace cpe
{


    void opengl_drawer::draw(const mesh& m)
    {
        draw_edges(m);
    }

    void opengl_drawer::draw(const mesh& m,const std::vector<double>& normal)
    {
        if(normal.size()==3*m.triangle_number())
            draw_per_polygon_normal(m,normal);
        else if(normal.size()==3*m.vertex_number())
            draw_per_vertex_normal(m,normal);
    }

    void opengl_drawer::draw_per_polygon_normal(const mesh& m,const std::vector<double>& normal)
    {
        const unsigned int* p_connectivity=&m.get_connectivity()[0];
        const double* p_vertex=&m.get_vertex()[0];
        const double* p_normal=&normal[0];
        const double* p_texture=&m.get_texture()[0];
        bool is_texture=(m.get_texture().size() == 2*m.vertex_number());




        glBegin(GL_TRIANGLES);

        unsigned int N_tri=m.triangle_number();

        if(is_texture==false)
        {
            for(unsigned int k_tri=0;k_tri<N_tri;++k_tri)
            {
                //set normal
                glNormal3dv(p_normal+3*k_tri);

                //draw vertices
                for(unsigned int k_dim=0;k_dim<3;++k_dim)
                    glVertex3dv(p_vertex+3*(*(p_connectivity++)));
            }
        }
        else
        {
            for(unsigned int k_tri=0;k_tri<N_tri;++k_tri)
            {
                //set normal
                glNormal3dv(p_normal+3*k_tri);

                //draw vertices
                for(unsigned int k_dim=0;k_dim<3;++k_dim)
                {
                    glTexCoord2dv(p_texture+2*(*(p_connectivity  )));
                    glVertex3dv  (p_vertex +3*(*(p_connectivity++)));
                }
            }
        }

        glEnd();
    }

    void opengl_drawer::draw_per_vertex_normal(const mesh& m,const std::vector<double>& normal)
    {
        unsigned int N_tri=m.triangle_number();
        const unsigned int* p_connectivity=&m.get_connectivity()[0];
        const double* p_vertex=&m.get_vertex()[0];
        const double* p_normal=&normal[0];
        const double* p_texture=&m.get_texture()[0];
        bool is_texture=(m.get_texture().size() == 2*m.vertex_number());


        //enable client state
        glEnableClientState(GL_VERTEX_ARRAY);
        glEnableClientState(GL_NORMAL_ARRAY);
        if(is_texture)
            glEnableClientState(GL_TEXTURE_COORD_ARRAY);

        //setup pointers
        glVertexPointer(3,GL_DOUBLE,0,p_vertex);
        glNormalPointer(GL_DOUBLE,0,p_normal);
        if(is_texture)
            glTexCoordPointer(2,GL_DOUBLE,0,p_texture);

        //draw triangles
        glDrawElements(GL_TRIANGLES,3*N_tri,GL_UNSIGNED_INT,p_connectivity);

        //disable client state
        glDisableClientState(GL_VERTEX_ARRAY);
        glDisableClientState(GL_NORMAL_ARRAY);
        if(is_texture)
            glDisableClientState(GL_TEXTURE_COORD_ARRAY);

    }

    void opengl_drawer::draw_edges(const mesh& m)
    {
        glDisable(GL_LIGHTING);

        const unsigned int* p_connectivity=&m.get_connectivity()[0];
        const double* p_vertex=&m.get_vertex()[0];


        unsigned int N_tri=m.triangle_number();

        //draw edge
        glBegin(GL_LINES);
        for(unsigned int k_tri=0;k_tri<N_tri;++k_tri)
        {
            for(unsigned int k_dim=0;k_dim<3;++k_dim)
            {
                glVertex3dv(p_vertex+3*p_connectivity[3*k_tri+k_dim]);
                glVertex3dv(p_vertex+3*p_connectivity[3*k_tri+(k_dim+1)%3]);
            }
        }
        glEnd();


    }



    void opengl_drawer::draw_normals(const mesh& m,const std::vector<double>& normal,const double& L_scaling)
    {
        glDisable(GL_LIGHTING);



        unsigned int N_tri=m.triangle_number();

        //draw per face normal
        if(normal.size()==3*m.triangle_number())
        {
            glBegin(GL_LINES);
            for(unsigned int k_tri=0;k_tri<N_tri;++k_tri)
            {
                //get barycenter of the triangle
                v3 v0=m.get_vertex(k_tri,0); v3 v1=m.get_vertex(k_tri,1); v3 v2=m.get_vertex(k_tri,2);
                v3 bar0 = 1.0/3.0*(v0+v1+v2);

                //get current normal
                v3 n(normal[3*k_tri+0],normal[3*k_tri+1],normal[3*k_tri+2]);

                //extremity
                v3 bar1=bar0+0.1*L_scaling*n;


                // draw normals
                glVertex3d(bar0.x(),bar0.y(),bar0.z());
                glVertex3d(bar1.x(),bar1.y(),bar1.z());

            }
            glEnd();
        }
        //draw per vertex normal
        else if(normal.size()==3*m.vertex_number())
        {
            const double* p_vertex=&m.get_vertex()[0];
            const double* p_normal=&normal[0];

            glBegin(GL_LINES);
            for(unsigned int k_v=0,N_v=m.vertex_number();k_v<N_v;++k_v)
            {


                //starting point
                glVertex3dv(p_vertex);
                //ending point
                glVertex3d(p_vertex[0]+0.1*L_scaling*p_normal[0],
                           p_vertex[1]+0.1*L_scaling*p_normal[1],
                           p_vertex[2]+0.1*L_scaling*p_normal[2]);

                p_vertex+=3;
                p_normal+=3;
            }
            glEnd();
        }



    }

    void opengl_drawer::generate_texture(const unsigned int& id_texture,const std::string& filename)
    {
        //load texture data
        texture t;
        t.load_picture(filename);

        GLuint id=id_texture;

        //generate texture id
        if(glIsTexture(id_texture))
            glDeleteTextures(1,&id);
        glGenTextures(1,&id);
        glBindTexture(GL_TEXTURE_2D,id);

        //build mipmap and store texture
        gluBuild2DMipmaps( GL_TEXTURE_2D, 4, t.get_Nx(),t.get_Ny(), GL_RGBA, GL_UNSIGNED_BYTE, &t.get_data()[0] );
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    }

}



